from datetime import datetime, timedelta, timezone
from typing import Optional

from werkzeug.exceptions import BadRequest, NotFound, Conflict
from sqlalchemy.exc import IntegrityError

from flaschengeist import logger
from flaschengeist.database import db
from flaschengeist.plugins.events import EventPlugin
from flaschengeist.plugins.events.models import EventType, Event, Job, JobType, JobTransfer, Service
from flaschengeist.utils.scheduler import scheduled


TRANSFER_REQUEST = 0x10
TRANSFER_ACCEPTED = 0x11
TRANSFER_REJECTED = 0x12


def update():
    db.session.commit()


def invite(job, invitee, sender):
    EventPlugin.plugin.notify(invitee, "Neue Diensteinladung", {"type": 0, "sender": sender.userid, "job": job.id})


def get_transfer(id):
    return JobTransfer.query.get(id)


def transfer(job: Job, new, old):
    service = Service.query.get((job.id, old.id_))
    if service is None:
        raise BadRequest

    jt = JobTransfer(old_user=old, new_user=new, job=job)
    db.session.add(jt)
    db.session.commit()
    EventPlugin.plugin.notify(
        new,
        "Neue Dienstübergabe",
        {"type": TRANSFER_REQUEST, "sender": old.userid, "event": job.event_id_, "job": job.id, "id": jt.id},
    )


def accept_transfer(jt: JobTransfer, accept=True):
    try:
        if accept:
            service = Service.query.get((jt.job.id, jt.old_user.id_))
            if service is not None:
                service.user_ = jt.new_user
            else:
                raise Conflict
            EventPlugin.plugin.notify(
                jt.old_user,
                "Dienstübergabe akzeptiert",
                {"type": TRANSFER_ACCEPTED, "sender": jt.new_user.userid, "event": jt.job.event_id_, "job": jt.job_id_},
            )
        else:
            EventPlugin.plugin.notify(
                jt.old_user,
                "Dienstübergabe abgelehnt",
                {"type": TRANSFER_REJECTED, "sender": jt.new_user.userid, "event": jt.job.event_id_, "job": jt.job_id_},
            )

    finally:
        db.session.delete(jt)
        db.session.commit()


def get_event_types():
    return EventType.query.all()


def get_event_type(identifier):
    """Get EventType by ID (int)"""

    if isinstance(identifier, int):
        et = EventType.query.get(identifier)
    else:
        logger.debug("Invalid identifier type for EventType")
        raise BadRequest
    if not et:
        raise NotFound
    return et


def create_event_type(name):
    try:
        event = EventType(name=name)
        db.session.add(event)
        db.session.commit()
        return event
    except IntegrityError:
        raise BadRequest("Name already exists")


def rename_event_type(identifier, new_name):
    event_type = get_event_type(identifier)
    event_type.name = new_name
    try:
        db.session.commit()
    except IntegrityError:
        raise BadRequest("Name already exists")


def delete_event_type(name):
    event_type = get_event_type(name)
    db.session.delete(event_type)
    try:
        db.session.commit()
    except IntegrityError:
        raise BadRequest("Type still in use")


def get_job_types():
    return JobType.query.all()


def get_job_type(type_id):
    job_type = JobType.query.get(type_id)
    print(job_type)
    if not job_type:
        raise NotFound
    return job_type


def create_job_type(name):
    try:
        job_type = JobType(name=name)
        db.session.add(job_type)
        db.session.commit()
        return job_type
    except IntegrityError:
        raise BadRequest("Name already exists")


def rename_job_type(name, new_name):
    job_type = get_job_type(name)
    job_type.name = new_name
    try:
        db.session.commit()
    except IntegrityError:
        raise BadRequest("Name already exists")


def delete_job_type(name):
    job_type = get_job_type(name)
    db.session.delete(job_type)
    try:
        db.session.commit()
    except IntegrityError:
        raise BadRequest("Type still in use")


def clear_backup(event: Event, backup):
    for job in event.jobs:
        services = []
        for service in job.services:
            if not service.is_backup or service.userid == backup:
                services.append(service)
        job.services = services


def get_event(event_id, with_backup=False) -> Event:
    event = Event.query.get(event_id)
    if event is None:
        raise NotFound
    return event


def get_templates():
    return Event.query.filter(Event.is_template == True).all()


def get_events(start: Optional[datetime] = None, end=None, with_backup=False):
    """Query events which start from begin until end
    Args:
        start (datetime): Earliest start
        end (datetime): Latest start
        with_backup (bool): Export also backup services

    Returns: collection of Event objects
    """
    query = Event.query.filter(Event.is_template.__eq__(False))
    if start is not None:
        query = query.filter(start <= Event.start)
    if end is not None:
        query = query.filter(Event.start < end)
    events = query.all()
    if not (with_backup is True):
        for event in events:
            clear_backup(event, with_backup)
    return events


def delete_event(event_id):
    """Delete event with given ID
    Args:
        event_id: id of Event to delete

    Raises:
        NotFound if not found
    """
    event = get_event(event_id)
    db.session.delete(event)
    db.session.commit()


def create_event(event_type, start, end=None, jobs=[], is_template=None, name=None, description=None):
    try:
        logger.debug(event_type)
        event = Event(
            start=start,
            end=end,
            name=name,
            description=description,
            type=event_type,
            is_template=is_template,
            jobs=jobs,
        )
        db.session.add(event)
        db.session.commit()
        return event
    except IntegrityError:
        logger.debug("Database error when creating new event", exc_info=True)
        raise BadRequest


def get_job(job_slot_id, event_id):
    js = Job.query.filter(Job.id == job_slot_id).filter(Job.event_id_ == event_id).one_or_none()
    if js is None:
        raise NotFound
    return js


def add_job(event, job_type, required_services, start, end=None, comment=None):
    job = Job(required_services=required_services, type=job_type, start=start, end=end, comment=comment)
    event.jobs.append(job)
    update()
    return job


def update():
    try:
        db.session.commit()
    except IntegrityError:
        logger.debug("Error, looks like a Job with that type already exists on an event", exc_info=True)
        raise BadRequest()


def delete_job(job: Job):
    db.session.delete(job)
    db.session.commit()


def assign_to_job(job: Job, user, value, backup=False):
    service = Service.query.get((job.id, user.id_))
    if value < 0:
        if not service:
            raise BadRequest
        db.session.delete(service)
    else:
        if service:
            service.value = value
            service.is_backup = backup
        else:
            service = Service(user_=user, value=value, job_=job, is_backup=backup)
            db.session.add(service)
    db.session.commit()


@scheduled
def assign_backups():
    logger.debug("Notifications")
    now = datetime.now(tz=timezone.utc)
    #       now +    backup_time      +  next cron tick
    start = now + timedelta(hours=16) + timedelta(minutes=30)
    services = Service.query.filter(Service.is_backup == True).join(Job).filter(Job.start <= start).all()
    for service in services:
        if service.job_.start <= now or service.job_.is_full():
            EventPlugin.plugin.notify(
                service.user_, "Your backup assignment was cancelled.", {"event_id": service.job_.event_id_}
            )
            logger.debug(f"Service is outdated or full, removing. {service.serialize()}")
            db.session.delete(service)
        else:
            service.is_backup = False
            logger.debug(f"Service not full, assigning backup. {service.serialize()}")
            EventPlugin.plugin.notify(
                service.user_, "Your backup assignment was accepted.", {"event_id": service.job_.event_id_}
            )
    db.session.commit()
